import logging
from threading import RLock
from pkg_resources import iter_entry_points

log = logging.getLogger(__name__)
lock = RLock()
EXTENSIONS = {}


def get_entry_points(section):
    """Load all Python classes registered at a given entry point."""
    with lock:
        if section not in EXTENSIONS:
            EXTENSIONS[section] = {}
            for ep in iter_entry_points(section):
                EXTENSIONS[section][ep.name] = ep.load()
        return EXTENSIONS[section]


def get_entry_point(section, name):
    return get_entry_points(section).get(name)


def get_extensions(section):
    """Iterate entry point objects."""
    return list(get_entry_points(section).values())



import logging
from datetime import datetime
from sqlalchemy import Column, DateTime, String
from sqlalchemy import Table, MetaData, JSON
from sqlalchemy import create_engine, func, select
from sqlalchemy.exc import IntegrityError
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.dialects.postgresql import insert as upsert

from servicelayer import settings

log = logging.getLogger(__name__)


class Tags(object):
    """Create persistent cached tags in a SQL database backend. This is
    a heavy-duty replacement for the simpler redis cache."""

    def __init__(self, name, uri=settings.TAGS_DATABASE_URI, **config):
        self.name = name
        self.engine = create_engine(uri, future=True, **config)
        self.is_postgres = self.engine.dialect.name == "postgresql"
        self.table = Table(
            name,
            MetaData(),
            Column("key", String, primary_key=True),  # noqa
            Column("value", JSONB if self.is_postgres else JSON),
            Column("timestamp", DateTime),
            extend_existing=True,
        )
        self.table.create(bind=self.engine, checkfirst=True)

    def delete(self, key=None, prefix=None):
        stmt = self.table.delete()
        if key is not None:
            stmt = stmt.where(self.table.c.key == key)
        if prefix is not None:
            stmt = stmt.where(self.table.c.key.startswith(prefix))
        with self.engine.connect() as conn:
            conn.execute(stmt)
            conn.commit()

    def close(self):
        self.engine.dispose()

    def get(self, key, since=None):
        stmt = select(self.table.c.value)
        stmt = stmt.where(self.table.c.key == key)
        if since is not None:
            stmt = stmt.where(self.table.c.timestamp >= since)
        with self.engine.connect() as conn:
            rp = conn.execute(stmt)
            row = rp.fetchone()
        if row is not None:
            return row.value

    def exists(self, key, since=None):
        stmt = select(func.count())
        stmt = stmt.where(self.table.c.key == key)
        if since is not None:
            stmt = stmt.where(self.table.c.timestamp >= since)
        with self.engine.connect() as conn:
            rp = conn.execute(stmt)
            count = rp.scalar()
        return count > 0

    def _store_values(self, conn, row):
        try:
            stmt = self.table.insert().values(row)
            conn.execute(stmt)
        except IntegrityError:
            changing = (
                "value",
                "timestamp",
            )
            changed = {c: row.get(c, {}) for c in changing}
            stmt = self.table.update().values(changed)
            stmt = stmt.where(self.table.c.key == row["key"])
            conn.execute(stmt)

    def _upsert_values(self, conn, row):
        """Use postgres' upsert mechanism (ON CONFLICT TO UPDATE)."""
        istmt = upsert(self.table).values(row)
        stmt = istmt.on_conflict_do_update(
            index_elements=["key"],
            set_=dict(
                value=istmt.excluded.value,
                timestamp=istmt.excluded.timestamp,
            ),
        )
        conn.execute(stmt)

    def set(self, key, value):
        conn = self.engine.connect()
        tx = conn.begin()
        now = datetime.utcnow()
        row = {"key": key, "value": value, "timestamp": now}
        try:
            if self.is_postgres:
                self._upsert_values(conn, row)
            else:
                self._store_values(conn, row)
            tx.commit()
        except Exception:
            tx.rollback()
            self.close()
            log.exception("Database error")

    def __repr__(self):
        return "<Tags(%r, %r)>" % (self.engine, self.name)



from unittest import TestCase
from datetime import datetime

from servicelayer.tags import Tags


class TagsTest(TestCase):
    def test_redis(self):
        tags = Tags("test_tags")
        assert "test_tags" in repr(tags)
        key = "banana"
        assert not tags.exists(key)
        assert not tags.get(key)
        tags.set(key, "split")
        assert tags.exists(key)
        assert tags.get(key) == "split", tags.get(key)
        tags.set(key, "new")
        assert tags.get(key) == "new", tags.get(key)
        now = datetime.utcnow()
        assert not tags.get(key, since=now)
        assert not tags.exists(key, since=now)
        tags.delete(key=key)
        assert not tags.get(key)
        assert not tags.exists(key)
        tags.set(key, "split")
        tags.delete()
        assert not tags.get(key)
        tags.set(key, "split")
        tags.delete(prefix="ba")
        assert not tags.get(key)
        tags.close()



import os
from hashlib import sha1
from pathlib import Path

BUF_SIZE = 1024 * 1024 * 16
HASH_LENGTH = 40  # sha1


def ensure_path(file_path):
    if file_path is None or isinstance(file_path, Path):
        return file_path
    return Path(file_path).resolve()


def ensure_posix_path(file_path):
    if isinstance(file_path, Path):
        file_path = file_path.as_posix()
    return file_path


def checksum(file_name):
    """Generate a hash for a given file name."""
    file_name = ensure_path(file_name)
    if file_name is not None and file_name.is_file():
        digest = sha1()
        with open(file_name, "rb") as fh:
            while True:
                block = fh.read(BUF_SIZE)
                if not block:
                    break
                digest.update(block)
        return str(digest.hexdigest())


def path_prefix(content_hash):
    """Get a prefix for a content hashed folder structure."""
    if content_hash is None:
        return None
    prefix = os.path.join(content_hash[:2], content_hash[2:4], content_hash[4:6])
    if len(content_hash) >= 6:
        prefix = os.path.join(prefix, content_hash)
    return prefix.rstrip(os.path.sep)


def path_content_hash(path):
    _, content_hash, _ = ensure_posix_path(path).rsplit(os.path.sep, 2)
    return content_hash
